/* This program takes a genbank flat file for a whole genome (i.e. a .gbk
 * file), and converts it to a format suitable for loading into a database.
 * The .gbk files in mind are specifically ones that contain full genomes, 
 * which is commonly found in genbank for archae. */

#include "common.h"
#include "hash.h"
#include "linefile.h"
#include "dystring.h"
#include "options.h"
#include "genbank.h"

#define FIELDSPACER '@'
/* This character is to replace slash characters beginning a line i.e. the */
/* sort that separate the fields from each other: */
/*        CDS       complement(52427..53416)
                    /gene="pdxA"
                    /locus_tag="b0052"
                    /function="enzyme/substrate cofactors, carriers:
                    Pyridoxine"
                    /note="pyridoxine biosynthesis;
                    go_component: cytoplasm [goid 0005737];
                    go_process: pyridoxine biosynthesis [goid 0008615]"
                    /codon_start=1                                        */
/* In this example, it should replace the slash in /gene, /locus_tag, */
/* /function, /note, and /codon_start, but NOT the slash in enzyme/substrate. */

#define DUMBSPACER '`'
/* This is for the problem this program was having of smashing two words together 
 * when processing the annotation sections of a genbank chunk and the sentence spanned
 * more than one line.  So in the output you'd get something like "proteinkinase" when 
 * "protein kinase" is desired.
 */

/* command line option specifications */
static struct optionSpec optionSpecs[] = {
    {"kind", OPTION_STRING},
    {NULL, 0}
};

char *parseRegions = NULL;
int lineCount = 1;  /* keyword lineNumber in parameter file outputs sequential line number */

void usage()
/* Explain usage and exit */
{
errAbort("gbArchaeGenome - Convert GenBank flat format file for whole genome (.gbk) to\n"
	 "tab-delimited based on parameters found in another file.\n"
         "usage:\n"
         "   gbArchaeGenome in.gbk param.txt out.tab\n"
         "options:\n"
	 "    -kind=cds   - Get CDS i.e. protein-coding genes.");
}

struct hash *hashUpChunk(char *chunk)
/* This will process a single chunk of the input, so it could be a gene, 
 * CDS, tRNA... whatever. It returns a hash of all the things
 * in the chunk.  Kind of an ugly function. */
{
struct hash *chunkHash = NULL;
char *parts[25], *moreparts[25], *locus, tmp[20];
int numparts, i;
unsigned s = 0, e = 0;
int len;
/* Create the hash with room for about 16 things initially */
/* (which should be more than enough, although the hash will */
/* automatically expand if needed. */
chunkHash = newHash(4);
/* Chop the whole thing up by the slashes. */
numparts = chopByChar(chunk, FIELDSPACER, parts, ArraySize(parts));
/* The first part should be the type/locus, and it needs to be chopped
 * by whitespace.  */
chopLine(parts[0], moreparts);
locus = cloneString(moreparts[1]);
if (differentWord(moreparts[0], parseRegions))
    return NULL;
hashAdd(chunkHash, "kind", cloneString(moreparts[0]));
hashAdd(chunkHash,"locus", locus);
tmp[0] = '+';
tmp[1] = '\0';
if (startsWith("complement", locus))
    tmp[0] = '-';
hashAdd(chunkHash, "strand", cloneString(tmp));
/* strip off mysterious back quote. */
/* This is in the locus-processing section.  I don't remember examples of */
/* this. */
len = strlen(locus);
if (locus[len-1] == DUMBSPACER)
    locus[len-1] = '\0';
/* parse that crazy locus part. */
genbankParseCds(locus, &s, &e);
safef(tmp, ArraySize(tmp), "%d", s);
hashAdd(chunkHash, "start", cloneString(tmp));
safef(tmp, ArraySize(tmp), "%d", e);
hashAdd(chunkHash, "end", cloneString(tmp));
for (i = 1; i < numparts; i++)
    {
    char *part = parts[i], *fieldAndVal[2];
    subChar(part, DUMBSPACER, ' '); 
    part = trimSpaces(part);
    stripChar(part, '\"');
    chopByChar(part, '=', fieldAndVal, 2);
    if (hashLookup(chunkHash, fieldAndVal[0]) == NULL)
        hashAdd(chunkHash, fieldAndVal[0], cloneString(trimSpaces(fieldAndVal[1])));
    else
        warn("Field repeated (%s %s)", fieldAndVal[0], trimSpaces(fieldAndVal[1]));
    }
return chunkHash;
}

void writeFields(struct hash *chunkHash, struct slName *paramList, FILE *output)
/* Write the fields to the file, based on the parameters. */
{
struct slName *current = paramList;
bool firstOne = TRUE;
while (current != NULL)
    /* Go through each parameter and get the VALUE part in the hash,
     * and for parameters not found in the table, just output the parameter. */
    {
    char *field = hashFindVal(chunkHash, current->name);
    if (field != NULL)
	{
	if (firstOne)
	    {
	    fprintf(output, "%s", field);
	    firstOne = FALSE;
	    }
	else
	    fprintf(output, "\t%s", field);
	}
    else
	{
	if (firstOne)
	    {
	    fprintf(output, "%s",current->name);
	    firstOne = FALSE;
	    }
	else
	    {
            if (atoi(current->name) != 0 || sameString(current->name, "0") 
                    || sameString(current->name, "cmpl") 
                    || startsWith("chr", current->name)
                    || startsWith("Contig", current->name))
             /* nummeric fields in parameter file are output without lookup */
             /* string cmpl is reserved for genePred format */
	        fprintf(output, "\t%s", current->name);
            else if (sameString(current->name, "lineNumber"))
             /* keyword lineNumber in parameter file outputs sequential line number */
                fprintf(output, "\t%d", lineCount++);
            else
            /* null value if keyword not found in genbank record */
                fprintf(output, "\t%s", "none");
	    }
	}
    current = current->next;
    }
fprintf(output, "\n");
}

void dealWithAChunk(char *chunk, struct slName *paramList, FILE *output)
/* Helper function of procGbFile to just convert a genbank chunk to the */
/* output we want. */
{
struct hash *chunkHash = hashUpChunk(chunk);
if (chunkHash != NULL)
    {
    writeFields(chunkHash, paramList, output);
    freeHashAndVals(&chunkHash);
    }
}

void procGbFile(char *inputName, struct slName *paramList, char *outputName)
/* Read and process a genbank file and deal with each desired chunk as they
 * are encountered. */
{
struct hash *tmp;
FILE *output = mustOpen(outputName, "w");
struct lineFile *lf = lineFileOpen(inputName, TRUE);
char *pieces[25];
char *line = NULL;
int wordCount;
struct dyString *chunk = NULL;
/* Read until the FEATURES part */
for (;;)
    {
    wordCount = lineFileChop(lf, pieces);
    if (sameString(pieces[0], "FEATURES"))
	break;
    }
/* Skip features line. */
lineFileSkip(lf,1);
/* Read in each gene/CDS part, and stop at the ORIGIN part. */
for (;;)
    {
    if (!lineFileNext(lf, &line, NULL))
	break;
    if (startsWith("ORIGIN", line) || startsWith("BASE COUNT", line))
	break;
    if (line[5] != ' ')
	{
	if (chunk != NULL)
	    {
	    dealWithAChunk(chunk->string, paramList, output);
	    dyStringClear(chunk);
	    }
	else
	    chunk = newDyString(256);
	dyStringAppend(chunk, line + 5);
	}
    else
	{
	if (chunk != NULL)
	    {
	    /* Change the */
	    if (line[21] == '/')
		line[21] = FIELDSPACER;
	    dyStringAppend(chunk, line + 21);	    
	    if (strlen(line) < 79)
		dyStringAppendC(chunk, DUMBSPACER);
	    }
	}
    }
if (chunk != NULL)
    dealWithAChunk(chunk->string, paramList, output);
freeDyString(&chunk);
lineFileClose(&lf);
carefulClose(&output);
}

struct slName* readParams(char *paramFile) 
/* Read the parameters file and store each as a node in a linked-list. */
{
struct slName *params = NULL;
struct lineFile *lf = lineFileOpen(paramFile, TRUE);
char *line = NULL;
for (;;)
    {
    if (!lineFileNext(lf,&line,NULL))
	break;
    slAddTail(&params, newSlName(trimSpaces(line)));
    }
lineFileClose(&lf);
return params;
}

int main(int argc, char *argv[])
/* Just the main... */
{
struct slName *paramList = NULL;
optionInit(&argc, argv, optionSpecs);
if (argc < 4)
    usage();
optionHash(&argc, argv);
parseRegions = optionVal("kind", "cds");
paramList = readParams(argv[2]);
procGbFile(argv[1], paramList, argv[3]);
slFreeList(&paramList);
return 0;
}
